import fm from 'front-matter';
import njk from 'nunjucks';
import path from 'path';
import fs from 'fs/promises';
import { Command } from 'commander';
import { collectItems } from '../utils/collectitems.mjs';
import express from 'express';
import { marked } from 'marked';
import * as puppeteer from 'puppeteer';

/**
 * @param prog {Command}
 */
export function addNjkCommand(prog) {
    prog.command('njk')
        .description('compile template into html')
        .argument('<filePath>', 'template file path')
        .option('--serve <port>', 'serve with express', 0)
        .option('--html <filePath>', 'save html', false)
        .option('--pdf <pdfPath>', 'save html as pdf', false)
        .option('--img <imgPath>', 'save html as img', false)
        .option('--scale <number>', 'scaling', 2)
        .action(async (str, option) => {
            try{
                if(option.serve){
                    const env = createNjkEnv(true);
                    const app = express();
                    app.get('/', async (req, res) => {
                        console.log(str);

                        try{
                            const html = await renderTemplate(env, str);
                            res.send(html);
                        }catch(e){
                            res.send(`<pre>error during template generation: ${e.stack}</pre>`);
                        }
                    });
                    console.log(`serving on ${option.serve}`);
                    app.listen(parseInt(option.serve));
                } else if (option.html) {
                    const env = createNjkEnv();
                    const html = await renderTemplate(env, str);

                    await fs.writeFile(option.html, html);
                } else if (option.pdf || option.img) {
                    const env = createNjkEnv();
                    const html = await renderTemplate(env, str);

                    await runPuppeteer(html, option.pdf, option.img, parseFloat(option.scale) || 2);
                } else {
                    const env = createNjkEnv();
                    console.log(await renderTemplate(env, str));
                }
            }catch(e){
                console.error(e);
                prog.error(`error during njk command`, { exitCode: 2, code: 'error-in-njk' });
            }
        });
}

/**
 * 
 * @param {boolean | undefined} watch 
 * @returns 
 */
export function createNjkEnv(watch) {
    const projdir = path.resolve(import.meta.url.slice(6), "../../..");
    const env = new njk.Environment([
        new njk.FileSystemLoader(['.', projdir], { watch }),
    ]);
    env.addFilter('md', txt => marked(txt));
    env.addFilter('icons8', txt => txt.replace(
        /:(([\w-]+\/)*[\w-]+):/g,
        '<img class="icon icons8" src="https://img.icons8.com/$1" />'
    ));
    return env;
}

/**
 * 
 * @param {njk.Environment} env 
 * @param {string} name 
 * @param {any} context 
 * @returns {Promise<string>}
 */
async function renderTemplate(env, name, context) {
    const attributes = await patchFrontmatter(env, name);
    const template = env.getTemplate(name);
    const children = template.render(context || attributes);

    if (typeof attributes.parent !== 'string') {
        return children;
    }else{
        context = { ...(context || attributes), children };
        return await renderTemplate(env, attributes.parent, context);
    }
}

/**
 * 
 * @param {njk.Environment} env
 * @param {str} name 
 * @returns {Promise<any>}
 */
async function patchFrontmatter(env, name) {
    const template = env.getTemplate(name);
    if (!template.fm) {
        const content = template.fm = fm(template.tmplStr);
        await collectItems(content);
        template.tmplStr = content.body;

        if(typeof content.attributes.parent === 'string'){
            content.attributes.parent = path.resolve(name, "..", content.attributes.parent);
            const frontParent = await patchFrontmatter(env, content.attributes.parent);
            content.attributes = { ...frontParent, ...content.attributes };
        }
    }
    return template.fm.attributes;
}


/**
 * 
 * @param {string} html 
 * @param {string} savePdf 
 * @param {string} saveImg 
 * @param {scale} deviceScaleFactor 
 */
async function runPuppeteer(html, savePdf, saveImg, scale=2){
    console.log(`launching puppeteer..., scale=${scale}`);
    const browser = await puppeteer.launch();

    console.log('loading page...');
    const page = await browser.newPage();
    await page.setContent(html);

    if(savePdf){
        console.log(`saving pdf to ${savePdf} ...`);
        await page.pdf({
            path: savePdf,
            printBackground: true,
            preferCSSPageSize: true,
        });
    }
    if(saveImg){
        console.log(`saving img to ${saveImg} ...`);
        await page.setViewport({
            width: 1024,
            height: 1024,
            deviceScaleFactor: scale,
        });
        await page.screenshot({
            path: saveImg,
            fullPage: true,
        });
    }

    console.log('closing browser...');
    await browser.close();
    console.log('all done!');
}