import { Command } from 'commander';
import fs from 'fs/promises';

/**
 * @param prog {Command}
 */
export function addPlotCommand(prog){
    prog.command('plot')
        .description('generates plot plt files')
        .argument('<plot-script>', 'js file that describes the plot')
        .option('-w, --width <number>', 'layout width', 287)
        .option('-h, --height <number>', 'layout height', 95)
        .option('-v, --verbose', 'prints debugs', false)
        .action(async (script, options) => {
            options.verbose && console.error(options);
            const code = await fs.readFile(script, { encoding: 'utf-8' });
            options.verbose && console.error(code);

            let plt = init(parseFloat(options.width), parseFloat(options.height));

            const GeneratorFunction = poly.constructor;
            /**
             * @type {(v: typeof pltCtx) => Iterable<unknown>}
             */
            const func = new GeneratorFunction('plt', code);
            for (const part of func(pltCtx)) {
                if (part[Symbol.iterator]) {
                    for (const cmd of poly(part))
                        plt += cmd;
                }else if(typeof part === 'string'){
                    plt += cmd;
                }
            }

            plt += end();

            console.log(plt);
        });
}

const pltCtx = {
    init,
    end,
    drad,
    plu,
    rrect,
    transform
}

/**
 * @param {number} w frame width
 * @param {number} h frame height
 * @returns {string} plt string
 */
function init(w,h){
    return ` IN TB26,${plu(w)},${plu(h)} CT1 U0,0 D0,0 D40,0`;
}

/**
 * @returns {string}
 */
function end() {
    return ' U0,0 @ @';
}

/**
 * 
 * @param {Iterable<[number,nu,ber]>} pts 
 * @param {(pt: [number, number]) => [number, number]} cb 
 */
function* transform(pts, cb) {
    for (const pt of pts)
        yield cb(pt);
}

/**
 * 
 * @param {number} x stx
 * @param {number} y sty
 * @param {number} w dx
 * @param {number} h dh
 * @param {number} r radius
 * @param {boolean} v verbose
 * @returns {Iterable<[number, number]>} rrect points
 */
function* rrect(x, y, w, h, r, v = false) {
    yield [x + w / 2, y];
    for (const [dx, dy] of drad(-90, 0, r)) {
        yield [x + w - r + dx, y + r + dy];
    }
    for (const [dx, dy] of drad(0, 90, r)) {
        yield [x + w - r + dx, y + h - r + dy];
    }
    for (const [dx, dy] of drad(90, 180, r)) {
        yield [x + r + dx, y + h - r + dy];
    }
    for (const [dx, dy] of drad(180, 270, r)) {
        yield [x + r + dx, y + r + dy];
    }
    yield [x + w / 2, y];
    v && console.error(`x=${x} y=${y} w=${w} h=${h} r=${r}`);
}

/**
 * 
 * @param {number} st start angle in degree
 * @param {number} ed end angle in degree
 * @param {number} radius
 * @returns {Iterable<[number, number]>}
 */
function* drad(st, ed, radius = 3) {
    const divides = Math.ceil(radius * 6);
    const segs = Math.ceil((ed - st) * divides / 360);
    for (let i = 0; i <= segs; i++) {
        const ang = st + (ed - st) / segs * i;
        const rad = ang / 180 * Math.PI;
        yield [Math.cos(rad) * radius, Math.sin(rad) * radius];
    }
}

/**
 * 
 * @param {Iterable<[number,number]>} pts points to cut.
 * @returns {Iterable<string>} plt commands
 */
function* poly(pts){
    let first = true;
    for(const pt of pts){
        yield ` ${first ? 'U' : 'D'}${plu(pt[0])},${plu(pt[1])}`;
        first = false;
    }
}

/**
 * 
 * @param {number} n measure in mm.
 * @returns {number} measure in plu (0.025mm).
 */
function plu(n) {
    return Math.round(n / 0.025);
}