import fm from "front-matter";
import _ from "lodash";
import { loadNotionDB } from "./notion.mjs";
import fetch from "node-fetch";
import { bootstrap } from "global-agent";
import { parse as csvParse } from "csv-parse";

bootstrap({
    environmentVariableNamespace: ''
});

/**
 * 
 * @param {fm.FrontMatterResult} content 
 */
export async function collectItems(content) {
    const { collectItems } = content.attributes;
    if (!collectItems) return;

    await Promise.all(collectItems.map(async item => {
        let collection = await loadCollection(item.src);
        if (item.duplicate) collection = duplicateCollection(collection, item.duplicate);
        if (item.map) collection = mapCollection(collection, item.map);
        content.attributes[item.key] = collection;
    }));
}

const loaded = {};
async function loadCollection(src) {
    if (!loaded[src]) {
        loaded[src] = await loadCollectionImpl(src);
    }
    return loaded[src];
}

/**
 *
 * @param {string} src
 * @return {Promise<any[]>}
 */
async function loadCollectionImpl(src) {
    if (src.match(/notion\.so/)) return loadNotionDB(src);

    const res = await fetch(src, { redirect: 'follow' });
    const contentType = res.headers.get('Content-Type');
    // console.error({ src, contentType });

    if(contentType.match(/text\/csv/)){
        const data = res.body.pipe(csvParse());
        const headers = [];
        const items = [];
        data.on('data', item => {
            if (headers.length) items.push(item);
            else headers.push(...item);
        });
        await new Promise((resolve, reject) => {
            data.on('end', resolve);
            data.on('error', reject);
        });
        return items.filter(body => {
            return body.find(item => !!item);
        }).map(body => {
            const obj = { };
            for(let i = 0; i < body.length; i ++){
                obj[headers[i]] = body[i];
            }

            return obj;
        });
    }
    if(contentType === 'application/json'){
        return res.json();
    }
    throw new Error(`unknown content type: ${contentType}`);
}

/**
 * 
 * @param {any[]} content
 * @param {string} key 
 * @returns {any[]}
 */
function duplicateCollection(content, key){
    return content.map(obj => {
        const n = parseInt(obj[key]) || 1;
        const all = [];
        for (let i = 0; i < n; i++)
            all.push(obj);
        return all;
    }).flat();
}

/**
 * 
 * @param {any[]} content
 * @param {string} key 
 * @returns {{[key: string]: any}}
 */
function mapCollection(content, key){
    const zamap = {};
    content.forEach(obj => {
        zamap[obj[key]] = obj;
    });
    return zamap;
}